/****************************************************************************
** Demonstration of using a custom keyboard handler to detect multiple     **
** cursor keys being held down simultaneously                              **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/
#include <dos.h>
#ifdef __DJGPP
  #include <go32.h>
  #include <dpmi.h>
#endif
#ifdef __cplusplus
  #include <conio.h>
  #include <stdio.h>
#endif

//Key status indicators
int CursorUp,
  CursorDown,
  CursorLeft,
  CursorRight;

//Pointer to original keyboard handler
#ifdef __DJGPP
  //and for DJGPP also to the new handler...
  _go32_dpmi_seginfo BIOSHandler, MyHandler;
#else
  #ifdef __cplusplus
    void interrupt (*BIOSHandler)(...);
  #else
    void interrupt (*BIOSHandler)();
  #endif
#endif

/****************************************************************************
** This is the custom keyboard handler. It checks to see if a cursor key   **
** is pressed or released and sets variables accordingly.                  **
****************************************************************************/

#ifdef __DJGPP
  void KeyCheck ()
#else
  #ifdef __cplusplus
    void interrupt KeyCheck (...)
  #else
    void interrupt KeyCheck ()
  #endif
#endif
{
  //Assume BIOS has to be called
  int CallBIOSHandler = 1;

  //Read scancode from keyboard
  unsigned char scanCode = inp (0x60);

  //Up key pressed
  if (scanCode == 72) {
    //Set up indicator
    CursorUp = 1;
    //No need to call BIOS
    CallBIOSHandler = 0;
  }
  //Up key depressed
  if (scanCode == 200) {
    //Reset up indicator
    CursorUp = 0;
    //No need to call BIOS
    CallBIOSHandler = 0;
  }
  //Left key pressed
  if (scanCode == 75) {
    //Set left indicator
    CursorLeft = 1;
    //No need to call BIOS
    CallBIOSHandler = 0;
  }
  //Left key depressed
  if (scanCode == 203) {
    //Reset left indicator
    CursorLeft = 0;
    //No need to call BIOS
    CallBIOSHandler = 0;
  }
  //Right key pressed
  if (scanCode == 77) {
    //Set right indicator
    CursorRight = 1;
    //No need to call BIOS
    CallBIOSHandler = 0;
  }
  //Right key depressed
  if (scanCode == 205) {
    //Reset right indicator
    CursorRight = 0;
    //No need to call BIOS
    CallBIOSHandler = 0;
  }
  //Down key pressed
  if (scanCode == 80) {
    //Set down indicator
    CursorDown = 1;
    //No need to call BIOS
    CallBIOSHandler = 0;
  }
  //Down key depressed
  if (scanCode == 208) {
    //Reset down indicator
    CursorDown = 0;
    //No need to call BIOS
    CallBIOSHandler = 0;
  }

  //If we haven't handled the keypress ourselves...
  if (CallBIOSHandler) {
    //...let the BIOS take care of it...
    #ifdef __DJGPP
      //DJGPP takes care of passing on to the BIOS iself
    #else
      asm { pushf }
      BIOSHandler ();
    #endif
  } else {
    //...otherwise, acknowledge the hardware interrupt
    outportb (0x20, 0x20);
  }
}

/****************************************************************************
** This procedures installs the custom keyboard handler and saves a        **
** pointer to the original                                                 **
****************************************************************************/
void XKey ()
{
  #ifdef __DJGPP
    /*
      Don't be scared now!
      If you're not using DJGPP, skip to the "#else" part!
    */

    //Get location of the new keyboard handler
    MyHandler.pm_offset = (int) KeyCheck;
    MyHandler.pm_selector = _go32_my_cs();
    //Save the old interrupt handler
    _go32_dpmi_get_protected_mode_interrupt_vector(9, &BIOSHandler);
    //Set the new interrupt handler in the chain
    _go32_dpmi_chain_protected_mode_interrupt_vector(9, &MyHandler);
  #else
    //Save BIOS keyboard handler
    BIOSHandler = getvect (9);
    //Set interrupt vector to our own routine
    setvect (9, KeyCheck);
  #endif
}

/****************************************************************************
** This procedures restores the original keyboard handler                  **
*****************************************************************************
** Make sure you call this before exiting your program! Otherwise, the     **
** system will crash                                                       **
****************************************************************************/
void NKey ()
{
  //Set interrupt vector to the BIOS keyboard handler
  #ifdef __DJGPP
    _go32_dpmi_set_protected_mode_interrupt_vector(9, &BIOSHandler);
  #else
    setvect (9, BIOSHandler);
  #endif
}

void main ()
{
  //Install eXtended KEYboardhandler
  XKey ();

  //Clear the screen
  clrscr ();

  //Show message
  printf ("Demonstration of custom keyboard handler\n");

  do {
    //Show cursor key status
    gotoxy (1, 3);
    printf (" Up    : %d\n", CursorUp);
    printf (" Down  : %d\n", CursorDown);
    printf (" Left  : %d\n", CursorLeft);
    printf (" Right : %d\n", CursorRight);
  } while (!kbhit() ? 1 : (getch() != 27));

  //Set Normal KEYboardhandler
  NKey ();
}